@@ -32,11 +32,16 @@
 use std::ffi::{CStr, CString, OsStr, OsString};
 use std::fmt;
 use std::mem;
-use std::io::{Result as IoResult, Error as IoError};
+use std::io;
 use std::os::unix::ffi::OsStrExt;
 use std::ptr;
 use std::sync::Arc;
 
+#[cfg(feature = "logging")]
+extern crate log;
+#[cfg(feature = "logging")]
+use self::log::debug;
+
 use libc::{c_char, uid_t, gid_t, c_int};
 use libc::passwd as c_passwd;
 use libc::group as c_group;
@@ -72,7 +77,7 @@ impl User {
         let name_arc = Arc::new(name.into());
         let extras = os::UserExtras::default();
 
-        User { uid, name_arc, primary_group, extras }
+        Self { uid, name_arc, primary_group, extras }
     }
 
     /// Returns this user’s ID.
@@ -186,7 +191,7 @@ impl Group {
         let name_arc = Arc::new(name.into());
         let extras = os::GroupExtras::default();
 
-        Group { gid, name_arc, extras }
+        Self { gid, name_arc, extras }
     }
 
     /// Returns this group’s ID.
@@ -244,34 +249,32 @@ unsafe fn from_raw_buf(p: *const c_char) -> OsString {
     OsStr::from_bytes(CStr::from_ptr(p).to_bytes()).to_os_string()
 }
 
-unsafe fn passwd_to_user(pointer: *const c_passwd) -> Option<User> {
-    if let Some(passwd) = pointer.as_ref().map(|p| ptr::read(p)) {
-        let name = Arc::new(from_raw_buf(passwd.pw_name));
+/// Reads data from the `c_passwd` and returns it as a `User`.
+unsafe fn passwd_to_user(passwd: c_passwd) -> User {
+    #[cfg(feature = "logging")]
+    debug!("Loading user with uid {}", passwd.pw_uid);
 
-        Some(User {
-            uid:           passwd.pw_uid,
-            name_arc:      name,
-            primary_group: passwd.pw_gid,
-            extras:        os::UserExtras::from_passwd(passwd),
-        })
-    }
-    else {
-        None
+    let name = Arc::new(from_raw_buf(passwd.pw_name));
+
+    User {
+        uid:           passwd.pw_uid,
+        name_arc:      name,
+        primary_group: passwd.pw_gid,
+        extras:        os::UserExtras::from_passwd(passwd),
     }
 }
 
-unsafe fn struct_to_group(pointer: *const c_group) -> Option<Group> {
-    if let Some(group) = pointer.as_ref().map(|p| ptr::read(p)) {
-        let name = Arc::new(from_raw_buf(group.gr_name));
+/// Reads data from the `c_group` and returns it as a `Group`.
+unsafe fn struct_to_group(group: c_group) -> Group {
+    #[cfg(feature = "logging")]
+    debug!("Loading group with gid {}", group.gr_gid);
 
-        Some(Group {
-            gid:       group.gr_gid,
-            name_arc:  name,
-            extras:    os::GroupExtras::from_struct(group),
-        })
-    }
-    else {
-        None
+    let name = Arc::new(from_raw_buf(group.gr_name));
+
+    Group {
+        gid:      group.gr_gid,
+        name_arc: name,
+        extras:   os::GroupExtras::from_struct(group),
     }
 }
 
@@ -322,6 +325,9 @@ pub fn get_user_by_uid(uid: uid_t) -> Option<User> {
     let mut buf = vec![0; 2048];  // TODO: Retry with larger buffer sizes
     let mut result = ptr::null_mut::<c_passwd>();
 
+    #[cfg(feature = "logging")]
+    debug!("Running getpwuid_r for user #{}", uid);
+
     unsafe {
         libc::getpwuid_r(uid, &mut passwd, buf.as_mut_ptr(), buf.len(), &mut result);
     }
@@ -337,7 +343,8 @@ pub fn get_user_by_uid(uid: uid_t) -> Option<User> {
         return None;
     }
 
-    unsafe { passwd_to_user(result) }
+    let user = unsafe { passwd_to_user(result.read()) };
+    Some(user)
 }
 
 /// Searches for a `User` with the given username in the system’s user database.
@@ -371,6 +378,9 @@ pub fn get_user_by_name<S: AsRef<OsStr> + ?Sized>(username: &S) -> Option<User>
     let mut buf = vec![0; 2048];  // TODO: Retry with larger buffer sizes
     let mut result = ptr::null_mut::<c_passwd>();
 
+    #[cfg(feature = "logging")]
+    debug!("Running getpwnam_r for user {:?}", username.as_ref());
+
     unsafe {
         libc::getpwnam_r(username.as_ptr(), &mut passwd, buf.as_mut_ptr(), buf.len(), &mut result);
     }
@@ -386,7 +396,8 @@ pub fn get_user_by_name<S: AsRef<OsStr> + ?Sized>(username: &S) -> Option<User>
         return None;
     }
 
-    unsafe { passwd_to_user(result) }
+    let user = unsafe { passwd_to_user(result.read()) };
+    Some(user)
 }
 
 /// Searches for a `Group` with the given ID in the system’s group database.
@@ -411,6 +422,9 @@ pub fn get_group_by_gid(gid: gid_t) -> Option<Group> {
     let mut buf = vec![0; 2048];  // TODO: Retry with larger buffer sizes
     let mut result = ptr::null_mut::<c_group>();
 
+    #[cfg(feature = "logging")]
+    debug!("Running getgruid_r for group #{}", gid);
+
     unsafe {
         libc::getgrgid_r(gid, &mut passwd, buf.as_mut_ptr(), buf.len(), &mut result);
     }
@@ -426,7 +440,8 @@ pub fn get_group_by_gid(gid: gid_t) -> Option<Group> {
         return None;
     }
 
-    unsafe { struct_to_group(result) }
+    let group = unsafe { struct_to_group(result.read()) };
+    Some(group)
 }
 
 /// Searches for a `Group` with the given group name in the system’s group database.
@@ -460,6 +475,9 @@ pub fn get_group_by_name<S: AsRef<OsStr> + ?Sized>(groupname: &S) -> Option<Grou
     let mut buf = vec![0; 2048];  // TODO: Retry with larger buffer sizes
     let mut result = ptr::null_mut::<c_group>();
 
+    #[cfg(feature = "logging")]
+    debug!("Running getgrnam_r for group {:?}", groupname.as_ref());
+
     unsafe {
         libc::getgrnam_r(groupname.as_ptr(), &mut group, buf.as_mut_ptr(), buf.len(), &mut result);
     }
@@ -475,7 +493,8 @@ pub fn get_group_by_name<S: AsRef<OsStr> + ?Sized>(groupname: &S) -> Option<Grou
         return None;
     }
 
-    unsafe { struct_to_group(result) }
+    let group = unsafe { struct_to_group(result.read()) };
+    Some(group)
 }
 
 /// Returns the user ID for the user running the process.
@@ -492,6 +511,9 @@ pub fn get_group_by_name<S: AsRef<OsStr> + ?Sized>(groupname: &S) -> Option<Grou
 /// println!("The ID of the current user is {}", get_current_uid());
 /// ```
 pub fn get_current_uid() -> uid_t {
+    #[cfg(feature = "logging")]
+    debug!("Running getuid");
+
     unsafe { libc::getuid() }
 }
 
@@ -534,6 +556,9 @@ pub fn get_current_username() -> Option<OsString> {
 /// println!("The ID of the effective user is {}", get_effective_uid());
 /// ```
 pub fn get_effective_uid() -> uid_t {
+    #[cfg(feature = "logging")]
+    debug!("Running geteuid");
+
     unsafe { libc::geteuid() }
 }
 
@@ -573,6 +598,9 @@ pub fn get_effective_username() -> Option<OsString> {
 /// println!("The ID of the current group is {}", get_current_gid());
 /// ```
 pub fn get_current_gid() -> gid_t {
+    #[cfg(feature = "logging")]
+    debug!("Running getgid");
+
     unsafe { libc::getgid() }
 }
 
@@ -612,6 +640,9 @@ pub fn get_current_groupname() -> Option<OsString> {
 /// println!("The ID of the effective group is {}", get_effective_gid());
 /// ```
 pub fn get_effective_gid() -> gid_t {
+    #[cfg(feature = "logging")]
+    debug!("Running getegid");
+
     unsafe { libc::getegid() }
 }
 
@@ -643,6 +674,11 @@ pub fn get_effective_groupname() -> Option<OsString> {
 ///
 /// - [`getgroups`](https://docs.rs/libc/*/libc/fn.getgroups.html)
 ///
+/// # Errors
+///
+/// This function will return `Err` when an I/O error occurs during the
+/// `getgroups` call.
+///
 /// # Examples
 ///
 /// ```no_run
@@ -652,20 +688,23 @@ pub fn get_effective_groupname() -> Option<OsString> {
 ///     println!("Process can access group #{} ({:?})", group.gid(), group.name());
 /// }
 /// ```
-pub fn group_access_list() -> IoResult<Vec<Group>> {
+pub fn group_access_list() -> io::Result<Vec<Group>> {
     let mut buff: Vec<gid_t> = vec![0; 1024];
 
+    #[cfg(feature = "logging")]
+    debug!("Running getgroups");
+
     let res = unsafe {
         libc::getgroups(1024, buff.as_mut_ptr())
     };
 
     if res < 0 {
-        Err(IoError::last_os_error())
+        Err(io::Error::last_os_error())
     }
     else {
         let mut groups = buff.into_iter()
-                                     .filter_map(|i| get_group_by_gid(i))
-                                     .collect::<Vec<_>>();
+                             .filter_map(get_group_by_gid)
+                             .collect::<Vec<_>>();
         groups.dedup_by_key(|i| i.gid());
         Ok(groups)
     }
@@ -696,6 +735,9 @@ pub fn get_user_groups<S: AsRef<OsStr> + ?Sized>(username: &S, gid: gid_t) -> Op
     let name = CString::new(username.as_ref().as_bytes()).unwrap();
     let mut count = buff.len() as c_int;
 
+    #[cfg(feature = "logging")]
+    debug!("Running getgrouplist for user {:?} and group #{}", username.as_ref(), gid);
+
     // MacOS uses i32 instead of gid_t in getgrouplist for unknown reasons
     #[cfg(all(unix, target_os="macos"))]
     let res = unsafe {
@@ -732,7 +774,7 @@ struct AllUsers;
 /// - [`setpwent`](https://docs.rs/libc/*/libc/fn.setpwent.html)
 /// - [`endpwent`](https://docs.rs/libc/*/libc/fn.endpwent.html)
 ///
-/// # Unsafety
+/// # Safety
 ///
 /// This constructor is marked as `unsafe`, which is odd for a crate
 /// that’s meant to be a safe interface. It *has* to be unsafe because
@@ -762,14 +804,25 @@ struct AllUsers;
 /// }
 /// ```
 pub unsafe fn all_users() -> impl Iterator<Item=User> {
+    #[cfg(feature = "logging")]
+    debug!("Running setpwent");
+
     #[cfg(not(target_os = "android"))]
     libc::setpwent();
     AllUsers
 }
 
 impl Drop for AllUsers {
+    #[cfg(target_os = "android")]
     fn drop(&mut self) {
-        #[cfg(not(target_os = "android"))]
+        // nothing to do here
+    }
+
+    #[cfg(not(target_os = "android"))]
+    fn drop(&mut self) {
+        #[cfg(feature = "logging")]
+        debug!("Running endpwent");
+
         unsafe { libc::endpwent() };
     }
 }
@@ -781,9 +834,21 @@ impl Iterator for AllUsers {
     fn next(&mut self) -> Option<User> {
         None
     }
+
     #[cfg(not(target_os = "android"))]
     fn next(&mut self) -> Option<User> {
-        unsafe { passwd_to_user(libc::getpwent()) }
+        #[cfg(feature = "logging")]
+        debug!("Running getpwent");
+
+        let result = unsafe { libc::getpwent() };
+
+        if result.is_null() {
+            None
+        }
+        else {
+            let user = unsafe { passwd_to_user(result.read()) };
+            Some(user)
+        }
     }
 }
 
@@ -873,7 +938,7 @@ pub mod os {
 
         impl Default for UserExtras {
             fn default() -> Self {
-                UserExtras {
+                Self {
                     home_dir: "/var/empty".into(),
                     shell:    "/bin/false".into(),
                     password: "*".into(),
@@ -884,7 +949,7 @@ pub mod os {
         impl UserExtras {
             /// Extract the OS-specific fields from the C `passwd` struct that
             /// we just read.
-            pub unsafe fn from_passwd(passwd: c_passwd) -> Self {
+            pub(crate) unsafe fn from_passwd(passwd: c_passwd) -> Self {
                 #[cfg(target_os = "android")]
                 {
                     Default::default()
@@ -895,7 +960,7 @@ pub mod os {
                     let shell    = from_raw_buf(passwd.pw_shell).into();
                     let password = from_raw_buf(passwd.pw_passwd);
 
-                    UserExtras { home_dir, shell, password }
+                    Self { home_dir, shell, password }
                 }
             }
         }
@@ -944,8 +1009,8 @@ pub mod os {
         impl GroupExtras {
             /// Extract the OS-specific fields from the C `group` struct that
             /// we just read.
-            pub unsafe fn from_struct(group: c_group) -> Self {
-                GroupExtras { members: members(group.gr_mem) }
+            pub(crate) unsafe fn from_struct(group: c_group) -> Self {
+                Self { members: members(group.gr_mem) }
             }
         }
 
@@ -990,8 +1055,8 @@ pub mod os {
         impl UserExtras {
             /// Extract the OS-specific fields from the C `passwd` struct that
             /// we just read.
-            pub unsafe fn from_passwd(passwd: c_passwd) -> Self {
-                UserExtras {
+            pub(crate) unsafe fn from_passwd(passwd: c_passwd) -> Self {
+                Self {
                     change: passwd.pw_change,
                     expire: passwd.pw_expire,
                     extras: super::unix::UserExtras::from_passwd(passwd),
@@ -1050,7 +1115,7 @@ pub mod os {
 
         impl Default for UserExtras {
             fn default() -> Self {
-                UserExtras {
+                Self {
                     extras: super::unix::UserExtras::default(),
                     change: 0,
                     expire: 0,
